==> jsonstreamreader-1.0.3.tar.gz: lib/jsontoken.dart <==
import 'dart:async';
import 'dart:math';

class JsonToken {
  static const int PAIR = 0;
  static const int VALUE = 1;
  static const int KEY = 2;
  static const int OBJECTSEPERATOR = 3;
  static const int ARRAYSEPERATOR = 4;
  static const int OBJECTSTART = 5;
  static const int OBJECTEND = 6;
  static const int ARRAYSTART = 7;
  static const int ARRAYEND = 8;
  static const int VALUEPAIR = 9;
  static const int NULL = 10;
  static Map<int, StreamController<List<MapEntry<String, dynamic>>>> _stream =
      new Map<int, StreamController<List<MapEntry<String, dynamic>>>>();
  static Map<int, Timer> _timer = new Map<int, Timer>();
  static Map<int, List<MapEntry<String, dynamic>>> _batch =
      new Map<int, List<MapEntry<String, dynamic>>>();
  dynamic value;
  int type;
  int index;
  int position;
  int parnterId;
  static Map<int, Map<int, JsonToken>> queue =
      new Map<int, Map<int, JsonToken>>();
  //unclosed objects ids
  static Map<int, int> _osi = new Map<int, int>();
  static Map<int, List<int>> os = new Map<int, List<int>>();
  //unclosed arrays ids
  static Map<int, int> _asi = new Map<int, int>();
  static Map<int, int> _fired = new Map<int, int>();
  static Map<int, List<int>> als = new Map<int, List<int>>();
  bool closed = false;
  String _key;
  static Map<int, List<JsonToken>> partners = new Map<int, List<JsonToken>>();
  static Map<int, List<JsonToken>> parents = new Map<int, List<JsonToken>>();
  static int pos = 0;
  JsonToken parent;
  JsonToken partner;
  JsonToken child;
  bool isKeyArray = false;
  static Map<int, String> _path = new Map<int, String>();
  static Map<int, int> _quotes = new Map<int, int>();
  int group;
  int _combine = 0;
  JsonToken(dynamic value, int type,
      {int position, int group, int combine = 0}) {
    this._combine = combine;
    if (group != null) {
      this.group = group;
      if (_batch[group] == null) {
        JsonToken.os[group] = new List<int>();
        JsonToken.als[group] = new List<int>();
        JsonToken.parents[group] = new List<JsonToken>();
        JsonToken.partners[group] = new List<JsonToken>();
        JsonToken._osi[group] = 0;
        JsonToken._asi[group] = 0;
        JsonToken._fired[group] = 0;
        JsonToken._path[group] = '/';
        JsonToken._quotes[group] = 0;
        _batch[group] = new List<MapEntry<String, dynamic>>();
        _stream[group] =
            new StreamController<List<MapEntry<String, dynamic>>>();
      }
    }
    this.value = value;
    this.type = type;
    this.group = group;
    this.position = JsonToken.pos;
    this._key = JsonToken._path[group].trim();
    JsonToken.pos++;
    //handle : , inside quotes
    if (this.value == '"') {
      JsonToken._quotes[group]++;
    }
    if ((JsonToken._quotes[group] % 2) == 1 &&
        (type == JsonToken.ARRAYSEPERATOR ||
            type == JsonToken.OBJECTSEPERATOR ||
            type == JsonToken.ARRAYSTART ||
            type == JsonToken.ARRAYEND ||
            type == JsonToken.OBJECTSTART ||
            type == JsonToken.OBJECTEND)) {
      this.type = JsonToken.VALUE;
    }

    if (JsonToken.queue[group] != null) {
      JsonToken.queue[group][(position != null ? position : this.position)] =
          this;
    } else {
      JsonToken.queue[group] = new Map<int, JsonToken>();
      JsonToken.queue[group][(position != null ? position : this.position)] =
          this;
    }
    if (this.type == JsonToken.OBJECTSTART) {
      //partner
      JsonToken._osi[group]++;
      JsonToken.os[group].add(JsonToken._osi[group]);
      this.parnterId = JsonToken._osi[group];
      //parent
      var parents = JsonToken.parents[group];
      if (parents.length > 0) {
        var parent = JsonToken.parents[group][parents.length - 1];
        this.parent = parent;
        parent.child = this;
      }
      JsonToken.parents[group].add(this);
    } else if (this.type == JsonToken.OBJECTEND) {
      this.parnterId = JsonToken.os[group].removeLast();
    } else if (this.type == JsonToken.ARRAYSTART) {
      //partner
      JsonToken._path[group] += '0/';
      JsonToken._asi[group]++;
      JsonToken.als[group].add(JsonToken._asi[group]);
      this.parnterId = JsonToken._asi[group];
      //parent
      var parents = JsonToken.parents[group];
      if (parents.length > 0) {
        var parent = JsonToken.parents[group][parents.length - 1];
        this.parent = parent;
        parent.child = this;
      }
      JsonToken.parents[group].add(this);
    } else if (this.type == JsonToken.ARRAYEND) {
      JsonToken._path[group] =
          JsonToken._path[group].replaceAll(new RegExp(r"[0-9]+\/$"), '');
      this.parnterId = JsonToken.als[group].removeLast();
    }
    //check for parents
    if (this.type == JsonToken.OBJECTSTART ||
        this.type == JsonToken.ARRAYSTART) {
      //these are parents
      JsonToken.partners[group].add(this);
    } else if (this.type == JsonToken.OBJECTEND ||
        this.type == JsonToken.ARRAYEND) {
      //these are parent closer
      this.partner = JsonToken.partners[group].removeLast();
      this.partner.closed = true;
      this.closed = true;
      //remove last parent
      JsonToken.parents[group].removeLast();
      this.parent = (JsonToken.parents[group].length == 0)
          ? null
          : JsonToken.parents[group][JsonToken.parents[group].length - 1];
    } else if (this.type == JsonToken.VALUE ||
        this.type == JsonToken.ARRAYSEPERATOR ||
        this.type == JsonToken.OBJECTSEPERATOR ||
        this.type == JsonToken.KEY) {
      //these are values or keys
      this.parent =
          JsonToken.partners[group][JsonToken.partners[group].length - 1];
    }
  }
  JsonToken append(JsonToken token) {
    JsonToken response;
    if (this.value.toString().endsWith("\\") && token.value == "\"") {
      JsonToken._quotes[group]--;
      this.value =
          this.value.toString().replaceAll(new RegExp(r"\\$"), "") + '\\"';
      token.value = null;
      token.type = JsonToken.NULL;
    }
    if (token.type == JsonToken.OBJECTSEPERATOR) {
      this.type = JsonToken.KEY;
    }
    if (this.type == JsonToken.OBJECTSEPERATOR &&
        token.type == JsonToken.VALUE) {
      token.type = JsonToken.VALUEPAIR;
    }
    if (token.type == JsonToken.OBJECTEND || token.type == JsonToken.ARRAYEND) {
      token.closed = true;
    }
    //release a simple token
    if ((token.type == JsonToken.ARRAYSEPERATOR ||
            token.type == JsonToken.ARRAYEND ||
            token.type == JsonToken.OBJECTEND) &&
        this.type == JsonToken.VALUE) {
      if (this.value == 'null') {
        this.value = null;
      } else if (this.value == 'true') {
        this.value = true;
      } else if (this.value == 'false') {
        this.value = false;
      }
      //release this token
      this.closed = true;
      if (this._combine == 1) {
        int lastIndex =
            int.parse(new RegExp(r"([0-9]+)\/$").firstMatch(this.key).group(1));
        String subkey = this.key.replaceAll(new RegExp(r"([0-9]+)\/$"), '');
        JsonToken.stream(
            this.group, subkey, {lastIndex.toString(): this.value});
      } else if (this._combine == 2) {
        String subkey = '';
        List<String> pieces = this.key.split('/');
        pieces.removeWhere((String piece) {
          return piece == null || piece == '';
        });
        if (pieces.length > 1) {
          pieces.getRange(1, pieces.length).toList().forEach((String k) {
            subkey += k + '/';
          });
        }
        JsonToken.stream(this.group, this.key.replaceFirst(subkey, ''),
            {'/' + subkey: this.value});
      } else {
        JsonToken.stream(this.group, this.key, this.value);
      }
      //current = new JsonToken(null, JsonToken.VALUE, this.position);
    }
    if (this.type == JsonToken.VALUE &&
        (token.type == JsonToken.VALUE || token.type == JsonToken.KEY)) {
      var r = token;
      this.value += r.value;
      this.value = this.value.replaceAll('&nbqts;', '\\"');
      token.value = null;
      token.type = JsonToken.NULL;
      response = this;
    } else if (this.type == JsonToken.VALUEPAIR &&
        token.type == JsonToken.VALUE) {
      var r = token;
      this.value += r.value;
      this.value = this.value.replaceAll('&nbqts;', '\\"');
      token.value = null;
      token.type = JsonToken.NULL;
      response = this;
    } else if (token.value == null) {
      response = this;
    } else {
      response = token;
    }
    if (this.type == JsonToken.OBJECTSEPERATOR &&
        token.type == JsonToken.ARRAYSTART) {
      token.isKeyArray = true;
    }
    //update key
    if (this.type == JsonToken.KEY) {
      JsonToken._path[this.group] +=
          this.value.replaceAll("\"", "").replaceAll("'", "").trim() + "/";
    }
    // or }} or ]}
    if ((this.type == JsonToken.OBJECTEND || this.type == JsonToken.ARRAYEND) &&
        token.type == JsonToken.OBJECTEND) {
      //remove last key
      List<String> pieces = JsonToken._path[this.group].split('/');
      pieces.removeLast();
      pieces.removeLast();
      JsonToken._path[this.group] = pieces.join('/') + '/';
      //update }
      token._key = JsonToken._path[this.group].trim();
    }
    //something like "f":[1,2,3]
    else if (this.type == JsonToken.ARRAYEND &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      if (this.partner.isKeyArray) {
        //remove last key
        List<String> pieces = JsonToken._path[this.group].split('/');
        pieces.removeLast();
        pieces.removeLast();
        JsonToken._path[this.group] = pieces.join('/') + '/';
      }
    }
    //:VALUEPAIR} or :VALUEPAIR,
    if (this.type == JsonToken.VALUEPAIR &&
        (token.type == JsonToken.OBJECTEND ||
            token.type == JsonToken.ARRAYSEPERATOR)) {
      this.closed = true;
      //remove last key
      List<String> pieces = JsonToken._path[this.group].split('/');
      pieces.removeLast();
      String k = pieces.removeLast();
      JsonToken._path[this.group] = pieces.join('/') + '/';
      //update }
      token._key = JsonToken._path[this.group].trim();
      JsonToken.stream(this.group, token._key, {k: this.value});
    }
    //also remove key for {}, and 0 parent is {
    if (this.type == JsonToken.OBJECTEND &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      this.closed = true;
      if (this.parent != null) {
        if (this.parent.type == JsonToken.OBJECTSTART) {
          //remove last key
          List<String> pieces = JsonToken._path[this.group].split('/');
          pieces.removeLast();
          pieces.removeLast();
          JsonToken._path[this.group] = pieces.join('/') + '/';
        }
      }
    }

    //also remove key for {}, or { }} with this.parent.type = {

    //[{},{}] or [a,b,c,d,e,f] or [[a,b,c],[a,b,c]]
    if ((this.type == JsonToken.VALUE ||
            this.type == JsonToken.OBJECTEND ||
            this.type == JsonToken.ARRAYEND) &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      this.closed = true;
      if (this.parent != null) {
        if (this.parent.type == JsonToken.ARRAYSTART) {
          int lastIndex = int.parse(new RegExp(r"([0-9]+)\/$")
              .firstMatch(JsonToken._path[this.group])
              .group(1));
          JsonToken._path[this.group] = JsonToken._path[this.group].replaceAll(
              new RegExp(r"([0-9]+)\/$"), (lastIndex + 1).toString() + '/');
        }
      }
    }
    return response;
  }

  String get key {
    return this._key;
  }

  toString() {
    return this.value;
  }

  static int getType(String value) {
    if (value == null) {
      return JsonToken.NULL;
    } else if (value == ']') {
      return JsonToken.ARRAYEND;
    } else if (value == '[') {
      return JsonToken.ARRAYSTART;
    } else if (value == '{') {
      return JsonToken.OBJECTSTART;
    } else if (value == '}') {
      return JsonToken.OBJECTEND;
    } else if (value == ',') {
      return JsonToken.ARRAYSEPERATOR;
    } else if (value == ':') {
      return JsonToken.OBJECTSEPERATOR;
    } else {
      return JsonToken.VALUE;
    }
  }

  static getFired(int group) {
    return JsonToken._fired[group];
  }

  static resetFired(int group) {
    return JsonToken._fired[group] = 0;
  }

  static stream(int group, String key, dynamic value) {
    JsonToken._fired[group]++;
    if (_batch[group].length > 0) {
      if (key == _batch[group].last.key) {
        if (_batch[group]
                .last
                .value
                .runtimeType
                .toString()
                .contains("HashMap") &&
            value.runtimeType.toString().contains("HashMap")) {
          _batch[group].last.value.addAll(value);
        } else {
          _batch[group].add(new MapEntry(key, value));
        }
      } else {
        _batch[group].add(new MapEntry(key, value));
      }
    } else {
      _batch[group].add(new MapEntry(key, value));
    }
    if (_timer[group] == null) {
      _timer[group] = new Timer(new Duration(microseconds: 1), () {
        _stream[group].add(_batch[group]);
        _timer[group] = null;
      });
    }
  }

  static identity() {
    var rng = new Random();
    return rng.nextInt(100000);
  }

  static close(int group) {
    _stream[group].close();
  }

  static Stream<List<MapEntry<String, dynamic>>> event(int group) {
    return _stream[group].stream;
  }
}

==> jsonstreamreader-1.1.1.tar.gz: lib/jsontoken.dart <==
import 'dart:async';
import 'dart:math';

import 'state/reader/token/jsontokenstate.dart';

//Token Produces by JsonTokenizer(Parser)
class JsonToken {
  static const int PAIR = 0;
  static const int VALUE = 1;
  static const int KEY = 2;
  static const int OBJECTSEPERATOR = 3;
  static const int ARRAYSEPERATOR = 4;
  static const int OBJECTSTART = 5;
  static const int OBJECTEND = 6;
  static const int ARRAYSTART = 7;
  static const int ARRAYEND = 8;
  static const int VALUEPAIR = 9;
  static const int NULL = 10;
  static Map<int, StreamController<List<MapEntry<String, dynamic>>>> _stream =
      new Map<int, StreamController<List<MapEntry<String, dynamic>>>>();
  static Map<int, Timer> _timer = new Map<int, Timer>();
  static Map<int, List<MapEntry<String, dynamic>>> _batch =
      new Map<int, List<MapEntry<String, dynamic>>>();
  dynamic value;
  int type;
  int index;
  int position;
  int parnterId;
  static Map<int, Map<int, JsonToken>> queue =
      new Map<int, Map<int, JsonToken>>();
  //unclosed objects ids
  static Map<int, int> _osi = new Map<int, int>();
  static Map<int, List<int>> os = new Map<int, List<int>>();
  //unclosed arrays ids
  static Map<int, int> _asi = new Map<int, int>();
  static Map<int, int> _fired = new Map<int, int>();
  static Map<int, List<int>> als = new Map<int, List<int>>();
  bool closed = false;
  String _key;
  static Map<int, List<JsonToken>> partners = new Map<int, List<JsonToken>>();
  static Map<int, List<JsonToken>> parents = new Map<int, List<JsonToken>>();
  static int pos = 0;
  JsonToken parent;
  JsonToken partner;
  JsonToken child;
  bool isKeyArray = false;
  static Map<int, String> _path = new Map<int, String>();
  static Map<int, int> _quotes = new Map<int, int>();
  int group;
  int _combine = 0;
  JsonToken.empty(
      dynamic statusvalue,
      String statuskey,
      int statustype,
      int statusposition,
      int statusgroup,
      int statuscombine,
      int statusparent,
      int statuspartner,
      int statuschild,
      bool statusisKey,
      int statusparnterId) {
    this.value = statusvalue;
    this._key = statuskey;
    this.type = statustype;
    this.position = statusposition;
    this.parnterId = statusparnterId;
    this.group = statusgroup;
    this._combine = statuscombine;
    this.isKeyArray = statusisKey;
    new Timer(new Duration(microseconds: 0), () {
      if (statusparent != null) {
        this.parent = JsonToken.queue[statusgroup][statusparent];
      }
      if (statuspartner != null) {
        this.partner = JsonToken.queue[statusgroup][statuspartner];
      }
      if (statuschild != null) {
        this.child = JsonToken.queue[statusgroup][statuschild];
      }
    });
  }
  JsonToken(dynamic value, int type,
      {int position, int group, int combine = 0}) {
    this._combine = combine;
    if (group != null) {
      this.group = group;
      if (_batch[group] == null) {
        JsonToken.os[group] = new List<int>();
        JsonToken.als[group] = new List<int>();
        JsonToken.parents[group] = new List<JsonToken>();
        JsonToken.partners[group] = new List<JsonToken>();
        JsonToken._osi[group] = 0;
        JsonToken._asi[group] = 0;
        JsonToken._fired[group] = 0;
        JsonToken._path[group] = '/';
        JsonToken._quotes[group] = 0;
        _batch[group] = new List<MapEntry<String, dynamic>>();
        _stream[group] =
            new StreamController<List<MapEntry<String, dynamic>>>();
      }
    }
    this.value = value;
    this.type = type;
    this.group = group;
    this.position = JsonToken.pos;
    this._key = JsonToken._path[group].trim();
    JsonToken.pos++;
    //handle : , inside quotes
    if (this.value == '"') {
      JsonToken._quotes[group]++;
    }
    if ((JsonToken._quotes[group] % 2) == 1 &&
        (type == JsonToken.ARRAYSEPERATOR ||
            type == JsonToken.OBJECTSEPERATOR ||
            type == JsonToken.ARRAYSTART ||
            type == JsonToken.ARRAYEND ||
            type == JsonToken.OBJECTSTART ||
            type == JsonToken.OBJECTEND)) {
      this.type = JsonToken.VALUE;
    }

    if (JsonToken.queue[group] != null) {
      JsonToken.queue[group][(position != null ? position : this.position)] =
          this;
    } else {
      JsonToken.queue[group] = new Map<int, JsonToken>();
      JsonToken.queue[group][(position != null ? position : this.position)] =
          this;
    }
    if (this.type == JsonToken.OBJECTSTART) {
      //partner
      JsonToken._osi[group]++;
      JsonToken.os[group].add(JsonToken._osi[group]);
      this.parnterId = JsonToken._osi[group];
      //parent
      var parents = JsonToken.parents[group];
      if (parents.length > 0) {
        var parent = JsonToken.parents[group][parents.length - 1];
        this.parent = parent;
        parent.child = this;
      }
      JsonToken.parents[group].add(this);
    } else if (this.type == JsonToken.OBJECTEND) {
      this.parnterId = JsonToken.os[group].removeLast();
    } else if (this.type == JsonToken.ARRAYSTART) {
      //partner
      JsonToken._path[group] += '0/';
      JsonToken._asi[group]++;
      JsonToken.als[group].add(JsonToken._asi[group]);
      this.parnterId = JsonToken._asi[group];
      //parent
      var parents = JsonToken.parents[group];
      if (parents.length > 0) {
        var parent = JsonToken.parents[group][parents.length - 1];
        this.parent = parent;
        parent.child = this;
      }
      JsonToken.parents[group].add(this);
    } else if (this.type == JsonToken.ARRAYEND) {
      JsonToken._path[group] =
          JsonToken._path[group].replaceAll(new RegExp(r"[0-9]+\/$"), '');
      this.parnterId = JsonToken.als[group].removeLast();
    }
    //check for parents
    if (this.type == JsonToken.OBJECTSTART ||
        this.type == JsonToken.ARRAYSTART) {
      //these are partners
      JsonToken.partners[group].add(this);
    } else if (this.type == JsonToken.OBJECTEND ||
        this.type == JsonToken.ARRAYEND) {
      //these are parent closer
      this.partner = JsonToken.partners[group].removeLast();
      this.partner.closed = true;
      this.closed = true;
      //remove last parent
      JsonToken.parents[group].removeLast();
      this.parent = (JsonToken.parents[group].length == 0)
          ? null
          : JsonToken.parents[group][JsonToken.parents[group].length - 1];
    } else if (this.type == JsonToken.VALUE ||
        this.type == JsonToken.ARRAYSEPERATOR ||
        this.type == JsonToken.OBJECTSEPERATOR ||
        this.type == JsonToken.KEY) {
      //these are values or keys
      this.parent =
          JsonToken.parents[group][JsonToken.parents[group].length - 1];
    }
  }
  static JsonTokenState getState(int group) {
    //unclosed arrays ids
    int stateasi = JsonToken._asi[group];
    int statefired = JsonToken._fired[group];
    //unclosed objects ids
    int stateosi = JsonToken._osi[group];
    String statepath = JsonToken._path[group];
    int statequotes = JsonToken._quotes[group];
    List<int> stateos = JsonToken.os[group];
    List<int> stateals = JsonToken.als[group];
    //
    Map<int, Map<String, dynamic>> statequeue =
        new Map<int, Map<String, dynamic>>();
    List<Map<String, dynamic>> statepartners = List<Map<String, dynamic>>();
    List<Map<String, dynamic>> stateparents = new List<Map<String, dynamic>>();

    JsonToken.queue[group].forEach((int key, JsonToken item) {
      statequeue[key] = {
        'position': item.position,
        'key': item.key,
        'value': item.value,
        'combine': item.combine,
        'group': item.group,
        'type': item.type,
        'parent': item.parent == null ? null : item.parent.position,
        'partner': item.partner == null ? null : item.partner.position,
        'child': item.child == null ? null : item.child.position,
        'isKeyArray': item.isKeyArray,
        'partnerId': item.parnterId
      };
    });
    JsonToken.partners[group].forEach((JsonToken item) {
      statepartners.add({
        'position': item.position,
        'key': item.key,
        'value': item.value,
        'combine': item.combine,
        'group': item.group,
        'type': item.type,
        'parent': item.parent == null ? null : item.parent.position,
        'partner': item.partner == null ? null : item.partner.position,
        'child': item.child == null ? null : item.child.position,
        'isKeyArray': item.isKeyArray,
        'partnerId': item.parnterId
      });
    });
    JsonToken.parents[group].forEach((JsonToken item) {
      stateparents.add({
        'position': item.position,
        'key': item.key,
        'value': item.value,
        'combine': item.combine,
        'group': item.group,
        'type': item.type,
        'parent': item.parent == null ? null : item.parent.position,
        'partner': item.partner == null ? null : item.partner.position,
        'child': item.child == null ? null : item.child.position,
        'isKeyArray': item.isKeyArray,
        'partnerId': item.parnterId
      });
    });
    return new JsonTokenState(
        statequeue,
        stateosi,
        stateos,
        stateasi,
        statefired,
        stateals,
        statepartners,
        stateparents,
        statepath,
        statequotes);
  }

  static void fromState(int groupnumber, JsonTokenState state) {
    //set JsonToken.pos to last position+1

    if (_batch[groupnumber] == null) {
      _batch[groupnumber] = new List<MapEntry<String, dynamic>>();
      _stream[groupnumber] =
          new StreamController<List<MapEntry<String, dynamic>>>();
      JsonToken.queue[groupnumber] = new Map<int, JsonToken>();
    } else {
      _batch[groupnumber] = new List<MapEntry<String, dynamic>>();
      _stream[groupnumber].close();
      _stream[groupnumber] =
          new StreamController<List<MapEntry<String, dynamic>>>();
      JsonToken.queue[groupnumber] = new Map<int, JsonToken>();
    }
    Map<String, dynamic> data = state.data();
    //unclosed arrays ids
    JsonToken._asi[groupnumber] = data['asi'];
    JsonToken._fired[groupnumber] = data['fired'];
    //unclosed objects ids
    JsonToken._osi[groupnumber] = data['osi'];
    JsonToken._path[groupnumber] = data['path'];
    JsonToken._quotes[groupnumber] = data['quotes'];
    JsonToken.os[groupnumber] = new List();
    JsonToken.os[groupnumber].addAll(data['os']);
    JsonToken.als[groupnumber] = new List();
    JsonToken.als[groupnumber].addAll(data['als']);
    //
    data['queue'].forEach((int key, Map<String, dynamic> item) {
      JsonToken.queue[groupnumber][key] = new JsonToken.empty(
          item['value'],
          item['key'],
          item['type'],
          item['position'],
          item['group'],
          item['combine'],
          item['parent'],
          item['partner'],
          item['child'],
          item['isKeyArray'],
          item['partnerId']);
      JsonToken.pos = item['position'] + 2;
    });
    JsonToken.parents[groupnumber] = new List();
    data['parents'].forEach((Map<String, dynamic> item) {
      JsonToken.parents[groupnumber].add(new JsonToken.empty(
          item['value'],
          item['key'],
          item['type'],
          item['position'],
          item['group'],
          item['combine'],
          item['parent'],
          item['partner'],
          item['child'],
          item['isKeyArray'],
          item['partnerId']));
    });
    JsonToken.partners[groupnumber] = new List();
    data['partners'].forEach((Map<String, dynamic> item) {
      JsonToken.partners[groupnumber].add(new JsonToken.empty(
          item['value'],
          item['key'],
          item['type'],
          item['position'],
          item['group'],
          item['combine'],
          item['parent'],
          item['partner'],
          item['child'],
          item['isKeyArray'],
          item['partnerId']));
    });
  }

  JsonToken append(JsonToken token) {
    JsonToken response;
    if (this.value.toString().endsWith("\\") && token.value == "\"") {
      JsonToken._quotes[group]--;
      this.value =
          this.value.toString().replaceAll(new RegExp(r"\\$"), "") + '\\"';
      token.value = null;
      token.type = JsonToken.NULL;
    }
    if (token.type == JsonToken.OBJECTSEPERATOR) {
      this.type = JsonToken.KEY;
    }
    if (this.type == JsonToken.OBJECTSEPERATOR &&
        token.type == JsonToken.VALUE) {
      token.type = JsonToken.VALUEPAIR;
    }
    if (token.type == JsonToken.OBJECTEND || token.type == JsonToken.ARRAYEND) {
      token.closed = true;
    }
    //release a simple token
    if ((token.type == JsonToken.ARRAYSEPERATOR ||
            token.type == JsonToken.ARRAYEND ||
            token.type == JsonToken.OBJECTEND) &&
        this.type == JsonToken.VALUE) {
      if (this.value == 'null') {
        this.value = null;
      } else if (this.value == 'true') {
        this.value = true;
      } else if (this.value == 'false') {
        this.value = false;
      }
      //release this token
      this.closed = true;
      if (this._combine == 1) {
        int lastIndex =
            int.parse(new RegExp(r"([0-9]+)\/$").firstMatch(this.key).group(1));
        String subkey = this.key.replaceAll(new RegExp(r"([0-9]+)\/$"), '');
        JsonToken.stream(
            this.group, subkey, {lastIndex.toString(): this.value});
      } else if (this._combine == 2) {
        String subkey = '';
        List<String> pieces = this.key.split('/');
        pieces.removeWhere((String piece) {
          return piece == null || piece == '';
        });
        if (pieces.length > 1) {
          pieces.getRange(1, pieces.length).toList().forEach((String k) {
            subkey += k + '/';
          });
        }
        JsonToken.stream(this.group, this.key.replaceFirst(subkey, ''),
            {'/' + subkey: this.value});
      } else {
        JsonToken.stream(this.group, this.key, this.value);
      }
      //current = new JsonToken(null, JsonToken.VALUE, this.position);
    }
    if (this.type == JsonToken.VALUE &&
        (token.type == JsonToken.VALUE || token.type == JsonToken.KEY)) {
      var r = token;
      this.value += r.value;
      this.value = this.value.replaceAll('&nbqts;', '\\"');
      token.value = null;
      token.type = JsonToken.NULL;
      response = this;
    } else if (this.type == JsonToken.VALUEPAIR &&
        token.type == JsonToken.VALUE) {
      var r = token;
      this.value += r.value;
      this.value = this.value.replaceAll('&nbqts;', '\\"');
      token.value = null;
      token.type = JsonToken.NULL;
      response = this;
    } else if (token.value == null) {
      response = this;
    } else {
      response = token;
    }
    if (this.type == JsonToken.OBJECTSEPERATOR &&
        token.type == JsonToken.ARRAYSTART) {
      token.isKeyArray = true;
    }
    //update key
    if (this.type == JsonToken.KEY) {
      JsonToken._path[this.group] +=
          this.value.replaceAll("\"", "").replaceAll("'", "").trim() + "/";
    }
    // or }} or ]}
    if ((this.type == JsonToken.OBJECTEND || this.type == JsonToken.ARRAYEND) &&
        token.type == JsonToken.OBJECTEND) {
      //remove last key
      List<String> pieces = JsonToken._path[this.group].split('/');
      pieces.removeLast();
      pieces.removeLast();
      JsonToken._path[this.group] = pieces.join('/') + '/';
      //update }
      token._key = JsonToken._path[this.group].trim();
    }
    //something like "f":[1,2,3]
    else if (this.type == JsonToken.ARRAYEND &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      if (this.partner.isKeyArray) {
        //remove last key
        List<String> pieces = JsonToken._path[this.group].split('/');
        pieces.removeLast();
        pieces.removeLast();
        JsonToken._path[this.group] = pieces.join('/') + '/';
      }
    }
    //:VALUEPAIR} or :VALUEPAIR,
    if (this.type == JsonToken.VALUEPAIR &&
        (token.type == JsonToken.OBJECTEND ||
            token.type == JsonToken.ARRAYSEPERATOR)) {
      this.closed = true;
      //remove last key
      List<String> pieces = JsonToken._path[this.group].split('/');
      pieces.removeLast();
      String k = pieces.removeLast();
      JsonToken._path[this.group] = pieces.join('/') + '/';
      //update }
      token._key = JsonToken._path[this.group].trim();
      JsonToken.stream(this.group, token._key, {k: this.value});
    }
    //also remove key for {}, and 0 parent is {
    if (this.type == JsonToken.OBJECTEND &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      this.closed = true;
      if (this.parent != null) {
        if (this.parent.type == JsonToken.OBJECTSTART) {
          //remove last key
          List<String> pieces = JsonToken._path[this.group].split('/');
          pieces.removeLast();
          pieces.removeLast();
          JsonToken._path[this.group] = pieces.join('/') + '/';
        }
      }
    }

    //also remove key for {}, or { }} with this.parent.type = {

    //[{},{}] or [a,b,c,d,e,f] or [[a,b,c],[a,b,c]]
    if ((this.type == JsonToken.VALUE ||
            this.type == JsonToken.OBJECTEND ||
            this.type == JsonToken.ARRAYEND) &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      this.closed = true;
      if (this.parent != null) {
        if (this.parent.type == JsonToken.ARRAYSTART) {
          int lastIndex = int.parse(new RegExp(r"([0-9]+)\/$")
              .firstMatch(JsonToken._path[this.group])
              .group(1));
          JsonToken._path[this.group] = JsonToken._path[this.group].replaceAll(
              new RegExp(r"([0-9]+)\/$"), (lastIndex + 1).toString() + '/');
        }
      }
    }
    return response;
  }

  String get key {
    return this._key;
  }

  int get combine {
    return this._combine;
  }

  toString() {
    return this.value;
  }

  static int getType(String value) {
    if (value == null) {
      return JsonToken.NULL;
    } else if (value == ']') {
      return JsonToken.ARRAYEND;
    } else if (value == '[') {
      return JsonToken.ARRAYSTART;
    } else if (value == '{') {
      return JsonToken.OBJECTSTART;
    } else if (value == '}') {
      return JsonToken.OBJECTEND;
    } else if (value == ',') {
      return JsonToken.ARRAYSEPERATOR;
    } else if (value == ':') {
      return JsonToken.OBJECTSEPERATOR;
    } else {
      return JsonToken.VALUE;
    }
  }

  static getFired(int group) {
    return JsonToken._fired[group];
  }

  static resetFired(int group) {
    return JsonToken._fired[group] = 0;
  }

  static stream(int group, String key, dynamic value) {
    JsonToken._fired[group]++;
    if (_batch[group].length > 0) {
      if (key == _batch[group].last.key) {
        if (_batch[group]
                .last
                .value
                .runtimeType
                .toString()
                .contains("HashMap") &&
            value.runtimeType.toString().contains("HashMap")) {
          _batch[group].last.value.addAll(value);
        } else {
          _batch[group].add(new MapEntry(key, value));
        }
      } else {
        _batch[group].add(new MapEntry(key, value));
      }
    } else {
      _batch[group].add(new MapEntry(key, value));
    }
    if (_timer[group] == null) {
      _timer[group] = new Timer(new Duration(microseconds: 1), () {
        _stream[group].add(_batch[group]);
        _timer[group] = null;
      });
    }
  }

  static identity() {
    var rng = new Random();
    return rng.nextInt(100000);
  }

  static close(int group) {
    _stream[group].close();
  }

  static Stream<List<MapEntry<String, dynamic>>> event(int group) {
    return _stream[group].stream;
  }
}
